package com.r3.corda.networkmanage.hsm.processor

import com.google.common.net.HostAndPort
import com.r3.corda.networkmanage.common.persistence.CrlIssuer
import com.r3.corda.networkmanage.common.persistence.PersistentCertificateRevocationListStorage
import com.r3.corda.networkmanage.common.persistence.PersistentCertificateRevocationRequestStorage
import com.r3.corda.networkmanage.common.persistence.RequestStatus
import com.r3.corda.networkmanage.common.signer.CertificateRevocationListSigner
import com.r3.corda.networkmanage.hsm.authentication.Authenticator
import com.r3.corda.networkmanage.hsm.authentication.createProvider
import com.r3.corda.networkmanage.hsm.configuration.DoormanCertificateConfig
import com.r3.corda.networkmanage.hsm.menu.Menu
import com.r3.corda.networkmanage.hsm.signer.HsmSigner
import com.r3.corda.networkmanage.hsm.utils.HsmX509Utilities.getAndInitializeKeyStore
import net.corda.nodeapi.internal.crypto.X509Utilities.CORDA_INTERMEDIATE_CA
import net.corda.nodeapi.internal.crypto.getX509Certificate
import net.corda.nodeapi.internal.persistence.CordaPersistence
import java.time.Duration

class CrrProcessor(private val config: DoormanCertificateConfig,
                   private val device: String,
                   private val keySpecifier: Int,
                   private val persistance: CordaPersistence) : Processor() {
    private companion object {
        private const val RESET = "\u001B[0m"
        private const val BLACK = "\u001B[30m"
        private const val RED = "\u001B[31m"
        private const val GREEN = "\u001B[32m"
        private const val YELLOW = "\u001B[33m"
        private const val BLUE = "\u001B[34m"
        private const val CYAN = "\u001B[36m"
        private const val WHITE = "\u001B[37m"
    }

    private val auth = config.authParameters

    fun showMenu() {
        val authenticator = Authenticator(
                provider = createProvider(config.keyGroup, keySpecifier, device),
                mode = auth.mode,
                authStrengthThreshold = auth.threshold)
        Menu().withExceptionHandler(this::processError).setExitOption("2", "Quit").addItem("1", "View current and sign a new certificate revocation list", {
            val crlStorage = PersistentCertificateRevocationListStorage(persistance)
            val currentCrl = crlStorage.getCertificateRevocationList(CrlIssuer.DOORMAN)
            printlnColor("Current CRL:")
            printlnColor(currentCrl.toString(), YELLOW)
            val crrStorage = PersistentCertificateRevocationRequestStorage(persistance)
            val approvedRequests = crrStorage.getRevocationRequests(RequestStatus.APPROVED)
            if (approvedRequests.isEmpty()) {
                printlnColor("There are no approved Certificate Revocation Requests.", GREEN)
            } else {
                printlnColor("Following are the approved Certificate Revocation Requests which will be added to the CRL:")
                approvedRequests.forEach {
                    printlnColor("Certificate DN: ${it.legalName}, Certificate serial number: ${it.certificateSerialNumber}", GREEN)
                }
            }
            Menu().withExceptionHandler(this::processError).setExitOption("2", "Go back").
                    addItem("1", "Create and sign a new Certificate Revocation List", {
                        authenticator.connectAndAuthenticate { provider, signers ->
                            val keyStore = getAndInitializeKeyStore(provider)
                            val issuerCertificate = keyStore.getX509Certificate(CORDA_INTERMEDIATE_CA)
                            val crlSigner = CertificateRevocationListSigner(
                                    revocationListStorage = crlStorage,
                                    issuerCertificate = issuerCertificate,
                                    updateInterval = Duration.ofMillis(config.crlUpdatePeriod),
                                    endpoint = config.crlDistributionPoint,
                                    signer = HsmSigner(provider = provider, keyName = CORDA_INTERMEDIATE_CA))
                            val currentRequests = crrStorage.getRevocationRequests(RequestStatus.DONE)
                            crlSigner.createSignedCRL(approvedRequests, currentRequests, signers.toString())
                        }
                    })
        }).showMenu()
    }
}