/*
 * ProGuard -- shrinking, optimization, obfuscation, and preverification
 *             of Java bytecode.
 *
 * Copyright (c) 2002-2017 Eric Lafortune @ GuardSquare
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package proguard.classfile.editor;

import proguard.classfile.*;
import proguard.classfile.attribute.*;
import proguard.classfile.attribute.visitor.AttributeVisitor;
import proguard.classfile.constant.Constant;
import proguard.classfile.instruction.*;
import proguard.classfile.util.*;

import static proguard.classfile.ClassConstants.*;

/**
 * This AttributeVisitor accumulates instructions and exceptions, and then
 * copies them into code attributes that it visits.
 *
 * @see CodeAttributeComposer
 *
 * @author Eric Lafortune
 */
public class CompactCodeAttributeComposer
extends      SimplifiedVisitor
implements   AttributeVisitor
{
    private final ConstantPoolEditor    constantPoolEditor;
    private final CodeAttributeComposer codeAttributeComposer;


    /**
     * Creates a new CompactCodeAttributeComposer that doesn't allow external
     * branch targets or exception handlers and that automatically shrinks
     * instructions.
     */
    public CompactCodeAttributeComposer(ProgramClass targetClass)
    {
        this(targetClass, false, false, true);
    }


    /**
     * Creates a new CompactCodeAttributeComposer.
     * @param allowExternalBranchTargets     specifies whether branch targets
     *                                       can lie outside the code fragment
     *                                       of the branch instructions.
     * @param allowExternalExceptionHandlers specifies whether exception
     *                                       handlers can lie outside the code
     *                                       fragment in which exceptions are
     *                                       defined.
     * @param shrinkInstructions             specifies whether instructions
     *                                       should automatically be shrunk
     *                                       before being written.
     */
    public CompactCodeAttributeComposer(ProgramClass targetClass,
                                        boolean      allowExternalBranchTargets,
                                        boolean      allowExternalExceptionHandlers,
                                        boolean      shrinkInstructions)
    {
        constantPoolEditor =
            new ConstantPoolEditor(targetClass);

        codeAttributeComposer =
            new CodeAttributeComposer(allowExternalBranchTargets,
                                      allowExternalExceptionHandlers,
                                      shrinkInstructions);
    }


    /**
     * Starts a new code definition.
     */
    public CompactCodeAttributeComposer reset()
    {
        codeAttributeComposer.reset();

        return this;
    }


    /**
     * Starts a new code fragment. Branch instructions that are added are
     * assumed to be relative within such code fragments.
     * @param maximumCodeFragmentLength the maximum length of the code that will
     *                                  be added as part of this fragment (more
     *                                  precisely, the maximum old instruction
     *                                  offset or label that is specified, plus
     *                                  one).
     */
    public CompactCodeAttributeComposer beginCodeFragment(int maximumCodeFragmentLength)
    {
        codeAttributeComposer.beginCodeFragment(maximumCodeFragmentLength);

        return this;
    }


    /**
     * Appends the given instruction with the given old offset.
     * Branch instructions must fit, for instance by enabling automatic
     * shrinking of instructions.
     * @param oldInstructionOffset the old offset of the instruction, to which
     *                             branches and other references in the current
     *                             code fragment are pointing.
     * @param instruction          the instruction to be appended.
     */
    public CompactCodeAttributeComposer appendInstruction(int         oldInstructionOffset,
                                                          Instruction instruction)
    {
        codeAttributeComposer.appendInstruction(oldInstructionOffset, instruction);

        return this;
    }


    /**
     * Appends the given label with the given old offset.
     * @param oldInstructionOffset the old offset of the label, to which
     *                             branches and other references in the current
     *                             code fragment are pointing.
     */
    public CompactCodeAttributeComposer appendLabel(int oldInstructionOffset)
    {
        codeAttributeComposer.appendLabel(oldInstructionOffset);

        return this;
    }


    /**
     * Appends the given instruction without defined offsets.
     * @param instructions the instructions to be appended.
     */
    public CompactCodeAttributeComposer appendInstructions(Instruction[] instructions)
    {
        codeAttributeComposer.appendInstructions(instructions);

        return this;
    }


    /**
     * Appends the given instruction without a defined offset.
     * Branch instructions should have a label, to allow computing the
     * new relative offset.
     * Branch instructions must fit, for instance by enabling automatic
     * shrinking of instructions.
     * @param instruction the instruction to be appended.
     */
    public CompactCodeAttributeComposer appendInstruction(Instruction instruction)
    {
        codeAttributeComposer.appendInstruction(instruction);

        return this;
    }


    /**
     * Appends the given exception to the exception table.
     * @param exceptionInfo the exception to be appended.
     */
    public CompactCodeAttributeComposer appendException(ExceptionInfo exceptionInfo)
    {
        codeAttributeComposer.appendException(exceptionInfo);

        return this;
    }


    /**
     * Inserts the given line number at the appropriate position in the line
     * number table.
     * @param lineNumberInfo the line number to be inserted.
     * @return the index where the line number was actually inserted.
     */
    public int insertLineNumber(LineNumberInfo lineNumberInfo)
    {
        return codeAttributeComposer.insertLineNumber(lineNumberInfo);
    }


    /**
     * Inserts the given line number at the appropriate position in the line
     * number table.
     * @param minimumIndex   the minimum index where the line number may be
     *                       inserted.
     * @param lineNumberInfo the line number to be inserted.
     * @return the index where the line number was inserted.
     */
    public int insertLineNumber(int minimumIndex, LineNumberInfo lineNumberInfo)
    {
        return codeAttributeComposer.insertLineNumber(minimumIndex, lineNumberInfo);
    }


    /**
     * Appends the given line number to the line number table.
     * @param lineNumberInfo the line number to be appended.
     */
    public CompactCodeAttributeComposer appendLineNumber(LineNumberInfo lineNumberInfo)
    {
        codeAttributeComposer.appendLineNumber(lineNumberInfo);

        return this;
    }


    /**
     * Wraps up the current code fragment, continuing with the previous one on
     * the stack.
     */
    public CompactCodeAttributeComposer endCodeFragment()
    {
        codeAttributeComposer.endCodeFragment();

        return this;
    }


    public CompactCodeAttributeComposer nop()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_NOP));
        return this;
    }

    public CompactCodeAttributeComposer aconst_null()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ACONST_NULL));
        return this;
    }

    public CompactCodeAttributeComposer iconst(int constant)
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_0, constant));
        return this;
    }

    public CompactCodeAttributeComposer iconst_m1()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_M1));
        return this;
    }

    public CompactCodeAttributeComposer iconst_0()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_0));
        return this;
    }

    public CompactCodeAttributeComposer iconst_1()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_1));
        return this;
    }

    public CompactCodeAttributeComposer iconst_2()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_2));
        return this;
    }

    public CompactCodeAttributeComposer iconst_3()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_3));
        return this;
    }

    public CompactCodeAttributeComposer iconst_4()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_4));
        return this;
    }

    public CompactCodeAttributeComposer iconst_5()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_5));
        return this;
    }

    public CompactCodeAttributeComposer lconst(int constant)
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LCONST_0, constant));
        return this;
    }

    public CompactCodeAttributeComposer lconst_0()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LCONST_0));
        return this;
    }

    public CompactCodeAttributeComposer lconst_1()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LCONST_1));
        return this;
    }

    public CompactCodeAttributeComposer fconst(int constant)
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCONST_0, constant));
        return this;
    }

    public CompactCodeAttributeComposer fconst_0()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCONST_0));
        return this;
    }

    public CompactCodeAttributeComposer fconst_1()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCONST_1));
        return this;
    }

    public CompactCodeAttributeComposer fconst_2()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCONST_2));
        return this;
    }

    public CompactCodeAttributeComposer dconst(int constant)
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCONST_0, constant));
        return this;
    }

    public CompactCodeAttributeComposer dconst_0()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCONST_0));
        return this;
    }

    public CompactCodeAttributeComposer dconst_1()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCONST_1));
        return this;
    }

    public CompactCodeAttributeComposer bipush(int constant)
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_BIPUSH, constant));
        return this;
    }

    public CompactCodeAttributeComposer sipush(int constant)
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_SIPUSH, constant));
        return this;
    }

    public CompactCodeAttributeComposer ldc(int value)
    {
        return ldc_(constantPoolEditor.addIntegerConstant(value));
    }

    public CompactCodeAttributeComposer ldc(float value)
    {
        return ldc_(constantPoolEditor.addFloatConstant(value));
    }

    public CompactCodeAttributeComposer ldc(String string)
    {
        return ldc(string, null, null);
    }

    public CompactCodeAttributeComposer ldc(String string, Clazz referencedClass, Method referencedMember)
    {
        return ldc_(constantPoolEditor.addStringConstant(string, referencedClass, referencedMember));
    }

    public CompactCodeAttributeComposer ldc(Clazz referencedClass)
    {
        return ldc(referencedClass.getName(), referencedClass);
    }

    public CompactCodeAttributeComposer ldc(String className, Clazz referencedClass)
    {
        return ldc_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactCodeAttributeComposer ldc_(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_LDC, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer ldc_w(int value)
    {
        return ldc_w_(constantPoolEditor.addIntegerConstant(value));
    }

    public CompactCodeAttributeComposer ldc_w(float value)
    {
        return ldc_w_(constantPoolEditor.addFloatConstant(value));
    }

    public CompactCodeAttributeComposer ldc_w(String string)
    {
        return ldc_w(string, null, null);
    }

    public CompactCodeAttributeComposer ldc_w(String string, Clazz referencedClass, Method referencedMember)
    {
        return ldc_w_(constantPoolEditor.addStringConstant(string, referencedClass, referencedMember));
    }

    public CompactCodeAttributeComposer ldc_w(String className, Clazz referencedClass)
    {
        return ldc_w_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactCodeAttributeComposer ldc_w_(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_LDC_W, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer ldc2_w(long value)
    {
        return ldc2_w(constantPoolEditor.addLongConstant(value));
    }

    public CompactCodeAttributeComposer ldc2_w(double value)
    {
        return ldc2_w(constantPoolEditor.addDoubleConstant(value));
    }

    public CompactCodeAttributeComposer ldc2_w(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_LDC2_W, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer iload(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer lload(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer fload(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer dload(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer aload(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer iload_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD_0));
        return this;
    }

    public CompactCodeAttributeComposer iload_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD_1));
        return this;
    }

    public CompactCodeAttributeComposer iload_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD_2));
        return this;
    }

    public CompactCodeAttributeComposer iload_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD_3));
        return this;
    }

    public CompactCodeAttributeComposer lload_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD_0));
        return this;
    }

    public CompactCodeAttributeComposer lload_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD_1));
        return this;
    }

    public CompactCodeAttributeComposer lload_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD_2));
        return this;
    }

    public CompactCodeAttributeComposer lload_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD_3));
        return this;
    }

    public CompactCodeAttributeComposer fload_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD_0));
        return this;
    }

    public CompactCodeAttributeComposer fload_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD_1));
        return this;
    }

    public CompactCodeAttributeComposer fload_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD_2));
        return this;
    }

    public CompactCodeAttributeComposer fload_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD_3));
        return this;
    }

    public CompactCodeAttributeComposer dload_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD_0));
        return this;
    }

    public CompactCodeAttributeComposer dload_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD_1));
        return this;
    }

    public CompactCodeAttributeComposer dload_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD_2));
        return this;
    }

    public CompactCodeAttributeComposer dload_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD_3));
        return this;
    }

    public CompactCodeAttributeComposer aload_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD_0));
        return this;
    }

    public CompactCodeAttributeComposer aload_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD_1));
        return this;
    }

    public CompactCodeAttributeComposer aload_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD_2));
        return this;
    }

    public CompactCodeAttributeComposer aload_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD_3));
        return this;
    }

    public CompactCodeAttributeComposer iaload()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IALOAD));
        return this;
    }

    public CompactCodeAttributeComposer laload()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LALOAD));
        return this;
    }

    public CompactCodeAttributeComposer faload()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FALOAD));
        return this;
    }

    public CompactCodeAttributeComposer daload()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DALOAD));
        return this;
    }

    public CompactCodeAttributeComposer aaload()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_AALOAD));
        return this;
    }

    public CompactCodeAttributeComposer baload()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_BALOAD));
        return this;
    }

    public CompactCodeAttributeComposer caload()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_CALOAD));
        return this;
    }

    public CompactCodeAttributeComposer saload()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_SALOAD));
        return this;
    }

    public CompactCodeAttributeComposer istore(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer lstore(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer fstore(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer dstore(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer astore(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer istore_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE_0));
        return this;
    }

    public CompactCodeAttributeComposer istore_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE_1));
        return this;
    }

    public CompactCodeAttributeComposer istore_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE_2));
        return this;
    }

    public CompactCodeAttributeComposer istore_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE_3));
        return this;
    }

    public CompactCodeAttributeComposer lstore_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE_0));
        return this;
    }

    public CompactCodeAttributeComposer lstore_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE_1));
        return this;
    }

    public CompactCodeAttributeComposer lstore_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE_2));
        return this;
    }

    public CompactCodeAttributeComposer lstore_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE_3));
        return this;
    }

    public CompactCodeAttributeComposer fstore_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE_0));
        return this;
    }

    public CompactCodeAttributeComposer fstore_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE_1));
        return this;
    }

    public CompactCodeAttributeComposer fstore_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE_2));
        return this;
    }

    public CompactCodeAttributeComposer fstore_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE_3));
        return this;
    }

    public CompactCodeAttributeComposer dstore_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE_0));
        return this;
    }

    public CompactCodeAttributeComposer dstore_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE_1));
        return this;
    }

    public CompactCodeAttributeComposer dstore_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE_2));
        return this;
    }

    public CompactCodeAttributeComposer dstore_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE_3));
        return this;
    }

    public CompactCodeAttributeComposer astore_0()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE_0));
        return this;
    }

    public CompactCodeAttributeComposer astore_1()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE_1));
        return this;
    }

    public CompactCodeAttributeComposer astore_2()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE_2));
        return this;
    }

    public CompactCodeAttributeComposer astore_3()
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE_3));
        return this;
    }

    public CompactCodeAttributeComposer iastore()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IASTORE));
        return this;
    }

    public CompactCodeAttributeComposer lastore()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LASTORE));
        return this;
    }

    public CompactCodeAttributeComposer fastore()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FASTORE));
        return this;
    }

    public CompactCodeAttributeComposer dastore()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DASTORE));
        return this;
    }

    public CompactCodeAttributeComposer aastore()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_AASTORE));
        return this;
    }

    public CompactCodeAttributeComposer bastore()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_BASTORE));
        return this;
    }

    public CompactCodeAttributeComposer castore()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_CASTORE));
        return this;
    }

    public CompactCodeAttributeComposer sastore()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_SASTORE));
        return this;
    }

    public CompactCodeAttributeComposer pop()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_POP));
        return this;
    }

    public CompactCodeAttributeComposer pop2()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_POP2));
        return this;
    }

    public CompactCodeAttributeComposer dup()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP));
        return this;
    }

    public CompactCodeAttributeComposer dup_x1()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP_X1));
        return this;
    }

    public CompactCodeAttributeComposer dup_x2()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP_X2));
        return this;
    }

    public CompactCodeAttributeComposer dup2()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP2));
        return this;
    }

    public CompactCodeAttributeComposer dup2_x1()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP2_X1));
        return this;
    }

    public CompactCodeAttributeComposer dup2_x2()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP2_X2));
        return this;
    }

    public CompactCodeAttributeComposer swap()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_SWAP));
        return this;
    }

    public CompactCodeAttributeComposer iadd()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IADD));
        return this;
    }

    public CompactCodeAttributeComposer ladd()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LADD));
        return this;
    }

    public CompactCodeAttributeComposer fadd()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FADD));
        return this;
    }

    public CompactCodeAttributeComposer dadd()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DADD));
        return this;
    }

    public CompactCodeAttributeComposer isub()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ISUB));
        return this;
    }

    public CompactCodeAttributeComposer lsub()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LSUB));
        return this;
    }

    public CompactCodeAttributeComposer fsub()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FSUB));
        return this;
    }

    public CompactCodeAttributeComposer dsub()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DSUB));
        return this;
    }

    public CompactCodeAttributeComposer imul()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IMUL));
        return this;
    }

    public CompactCodeAttributeComposer lmul()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LMUL));
        return this;
    }

    public CompactCodeAttributeComposer fmul()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FMUL));
        return this;
    }

    public CompactCodeAttributeComposer dmul()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DMUL));
        return this;
    }

    public CompactCodeAttributeComposer idiv()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IDIV));
        return this;
    }

    public CompactCodeAttributeComposer ldiv()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LDIV));
        return this;
    }

    public CompactCodeAttributeComposer fdiv()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FDIV));
        return this;
    }

    public CompactCodeAttributeComposer ddiv()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DDIV));
        return this;
    }

    public CompactCodeAttributeComposer irem()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IREM));
        return this;
    }

    public CompactCodeAttributeComposer lrem()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LREM));
        return this;
    }

    public CompactCodeAttributeComposer frem()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FREM));
        return this;
    }

    public CompactCodeAttributeComposer drem()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DREM));
        return this;
    }

    public CompactCodeAttributeComposer ineg()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_INEG));
        return this;
    }

    public CompactCodeAttributeComposer lneg()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LNEG));
        return this;
    }

    public CompactCodeAttributeComposer fneg()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FNEG));
        return this;
    }

    public CompactCodeAttributeComposer dneg()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DNEG));
        return this;
    }

    public CompactCodeAttributeComposer ishl()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ISHL));
        return this;
    }

    public CompactCodeAttributeComposer lshl()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LSHL));
        return this;
    }

    public CompactCodeAttributeComposer ishr()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ISHR));
        return this;
    }

    public CompactCodeAttributeComposer lshr()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LSHR));
        return this;
    }

    public CompactCodeAttributeComposer iushr()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IUSHR));
        return this;
    }

    public CompactCodeAttributeComposer lushr()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LUSHR));
        return this;
    }

    public CompactCodeAttributeComposer iand()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IAND));
        return this;
    }

    public CompactCodeAttributeComposer land()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LAND));
        return this;
    }

    public CompactCodeAttributeComposer ior()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IOR));
        return this;
    }

    public CompactCodeAttributeComposer lor()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LOR));
        return this;
    }

    public CompactCodeAttributeComposer ixor()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IXOR));
        return this;
    }

    public CompactCodeAttributeComposer lxor()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LXOR));
        return this;
    }

    public CompactCodeAttributeComposer iinc(int variableIndex,
                                             int constant)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_IINC, variableIndex, constant));
        return this;
    }

    public CompactCodeAttributeComposer i2l()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2L));
        return this;
    }

    public CompactCodeAttributeComposer i2f()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2F));
        return this;
    }

    public CompactCodeAttributeComposer i2d()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2D));
        return this;
    }

    public CompactCodeAttributeComposer l2i()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_L2I));
        return this;
    }

    public CompactCodeAttributeComposer l2f()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_L2F));
        return this;
    }

    public CompactCodeAttributeComposer l2d()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_L2D));
        return this;
    }

    public CompactCodeAttributeComposer f2i()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_F2I));
        return this;
    }

    public CompactCodeAttributeComposer f2l()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_F2L));
        return this;
    }

    public CompactCodeAttributeComposer f2d()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_F2D));
        return this;
    }

    public CompactCodeAttributeComposer d2i()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_D2I));
        return this;
    }

    public CompactCodeAttributeComposer d2l()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_D2L));
        return this;
    }

    public CompactCodeAttributeComposer d2f()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_D2F));
        return this;
    }

    public CompactCodeAttributeComposer i2b()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2B));
        return this;
    }

    public CompactCodeAttributeComposer i2c()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2C));
        return this;
    }

    public CompactCodeAttributeComposer i2s()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2S));
        return this;
    }

    public CompactCodeAttributeComposer lcmp()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LCMP));
        return this;
    }

    public CompactCodeAttributeComposer fcmpl()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCMPL));
        return this;
    }

    public CompactCodeAttributeComposer fcmpg()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCMPG));
        return this;
    }

    public CompactCodeAttributeComposer dcmpl()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCMPL));
        return this;
    }

    public CompactCodeAttributeComposer dcmpg()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCMPG));
        return this;
    }

    public CompactCodeAttributeComposer ifeq(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFEQ, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ifne(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFNE, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer iflt(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFLT, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ifge(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFGE, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ifgt(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFGT, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ifle(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFLE, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ificmpeq(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPEQ, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ificmpne(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPNE, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ificmplt(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPLT, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ificmpge(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPGE, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ificmpgt(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPGT, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ificmple(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPLE, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ifacmpeq(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFACMPEQ, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ifacmpne(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFACMPNE, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer goto_(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_GOTO, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer jsr(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_JSR, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ret(int variableIndex)
    {
        codeAttributeComposer.appendInstruction(new VariableInstruction(InstructionConstants.OP_RET, variableIndex));
        return this;
    }

    public CompactCodeAttributeComposer tableswitch(int   defaultOffset,
                                                    int   lowCase,
                                                    int   highCase,
                                                    int[] jumpOffsets)
    {
        codeAttributeComposer.appendInstruction(new TableSwitchInstruction(InstructionConstants.OP_TABLESWITCH,
                                                                           defaultOffset,
                                                                           lowCase,
                                                                           highCase,
                                                                           jumpOffsets));
        return this;
    }

    public CompactCodeAttributeComposer lookupswitch(int  defaultOffset,
                                                    int[] cases,
                                                    int[] jumpOffsets)
    {
        codeAttributeComposer.appendInstruction(new LookUpSwitchInstruction(InstructionConstants.OP_LOOKUPSWITCH,
                                                                            defaultOffset,
                                                                            cases,
                                                                            jumpOffsets));
        return this;
    }

    public CompactCodeAttributeComposer ireturn()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_IRETURN));
        return this;
    }

    public CompactCodeAttributeComposer lreturn()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_LRETURN));
        return this;
    }

    public CompactCodeAttributeComposer freturn()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_FRETURN));
        return this;
    }

    public CompactCodeAttributeComposer dreturn()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_DRETURN));
        return this;
    }

    public CompactCodeAttributeComposer areturn()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ARETURN));
        return this;
    }

    public CompactCodeAttributeComposer return_()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_RETURN));
        return this;
    }

    /**
     * Appends the proper return statement for the given internal type.
     *
     * @param internalType  the return type.
     */
    public void return_(String internalType)
    {
        switch (internalType.charAt(0))
        {
            case TYPE_BOOLEAN:
            case TYPE_BYTE:
            case TYPE_CHAR:
            case TYPE_SHORT:
            case TYPE_INT:
                ireturn();
                break;

            case TYPE_LONG:
                lreturn();
                break;

            case TYPE_FLOAT:
                freturn();
                break;

            case TYPE_DOUBLE:
                dreturn();
                break;

            case TYPE_VOID:
                return_();
                break;

            case TYPE_CLASS_START:
            case TYPE_ARRAY:
                areturn();
                break;
        }
    }

    public CompactCodeAttributeComposer getstatic(Clazz  referencedClass,
                                                  Member referencedMember)
    {
        return getstatic(referencedClass.getName(),
                         referencedMember.getName(referencedClass),
                         referencedMember.getDescriptor(referencedClass),
                         referencedClass,
                         referencedMember);
    }

    public CompactCodeAttributeComposer getstatic(String className,
                                                  String name,
                                                  String descriptor)
    {
        return getstatic(className, name, descriptor, null, null);
    }

    public CompactCodeAttributeComposer getstatic(String className,
                                                  String name,
                                                  String descriptor,
                                                  Clazz  referencedClass,
                                                  Member referencedMember)
    {
        return getstatic(constantPoolEditor.addFieldrefConstant(className,
                                                                name,
                                                                descriptor,
                                                                referencedClass,
                                                                referencedMember));
    }

    public CompactCodeAttributeComposer getstatic(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_GETSTATIC, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer putstatic(String className,
                                                  String name,
                                                  String descriptor)
    {
        return putstatic(className, name, descriptor, null, null);
    }

    public CompactCodeAttributeComposer putstatic(Clazz  referencedClass,
                                                  Member referencedMember)
    {
        return putstatic(referencedClass.getName(),
                         referencedMember.getName(referencedClass),
                         referencedMember.getDescriptor(referencedClass),
                         referencedClass,
                         referencedMember);
    }

    public CompactCodeAttributeComposer putstatic(String className,
                                                  String name,
                                                  String descriptor,
                                                  Clazz  referencedClass,
                                                  Member referencedMember)
    {
        return putstatic(constantPoolEditor.addFieldrefConstant(className,
                                                                name,
                                                                descriptor,
                                                                referencedClass,
                                                                referencedMember));
    }

    public CompactCodeAttributeComposer putstatic(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_PUTSTATIC, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer getfield(String className,
                                                 String name,
                                                 String descriptor)
    {
        return getfield(className, name, descriptor, null, null);
    }

    public CompactCodeAttributeComposer getfield(String className,
                                                 String name,
                                                 String descriptor,
                                                 Clazz  referencedClass,
                                                 Member referencedMember)
    {
        return getfield(constantPoolEditor.addFieldrefConstant(className,
                                                               name,
                                                               descriptor,
                                                               referencedClass,
                                                               referencedMember));
    }

    public CompactCodeAttributeComposer getfield(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_GETFIELD, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer putfield(String className,
                                                 String name,
                                                 String descriptor)
    {
        return putfield(className, name, descriptor, null, null);
    }

    public CompactCodeAttributeComposer putfield(String className,
                                                 String name,
                                                 String descriptor,
                                                 Clazz  referencedClass,
                                                 Member referencedMember)
    {
        return putfield(constantPoolEditor.addFieldrefConstant(className,
                                                               name,
                                                               descriptor,
                                                               referencedClass,
                                                               referencedMember));
    }

    public CompactCodeAttributeComposer putfield(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_PUTFIELD, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer invokevirtual(String className,
                                                      String name,
                                                      String descriptor)
    {
        return invokevirtual(className, name, descriptor, null, null);
    }

    public CompactCodeAttributeComposer invokevirtual(String className,
                                                      String name,
                                                      String descriptor,
                                                      Clazz  referencedClass,
                                                      Member referencedMember)
    {
        return invokevirtual(constantPoolEditor.addMethodrefConstant(className,
                                                                     name,
                                                                     descriptor,
                                                                     referencedClass,
                                                                     referencedMember));
    }

    public CompactCodeAttributeComposer invokevirtual(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKEVIRTUAL, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer invokespecial(String className,
                                                      String name,
                                                      String descriptor)
    {
        return invokespecial(className, name, descriptor, null, null);
    }

    public CompactCodeAttributeComposer invokespecial(String className,
                                                      String name,
                                                      String descriptor,
                                                      Clazz  referencedClass,
                                                      Member referencedMember)
    {
        return invokespecial(constantPoolEditor.addMethodrefConstant(className,
                                                                     name,
                                                                     descriptor,
                                                                     referencedClass,
                                                                     referencedMember));
    }

    public CompactCodeAttributeComposer invokespecial(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKESPECIAL, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer invokestatic(String className,
                                                     String name,
                                                     String descriptor)
    {
        return invokestatic(className, name, descriptor, null, null);
    }

    public CompactCodeAttributeComposer invokestatic(String className,
                                                     String name,
                                                     String descriptor,
                                                     Clazz  referencedClass,
                                                     Member referencedMember)
    {
        return invokestatic(constantPoolEditor.addMethodrefConstant(className,
                                                                    name,
                                                                    descriptor,
                                                                    referencedClass,
                                                                    referencedMember));
    }

    public CompactCodeAttributeComposer invokestatic(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKESTATIC, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer invokestaticinterface(String className,
                                                              String name,
                                                              String descriptor,
                                                              Clazz  referencedClass,
                                                              Member referencedMember)
    {
        return invokestatic(constantPoolEditor.addInterfaceMethodrefConstant(className,
                                                                             name,
                                                                             descriptor,
                                                                             referencedClass,
                                                                             referencedMember));
    }

    public CompactCodeAttributeComposer invokeinterface(String className,
                                                        String name,
                                                        String descriptor)
    {
        return invokeinterface(className, name, descriptor, null, null);
    }

    public CompactCodeAttributeComposer invokeinterface(String className,
                                                        String name,
                                                        String descriptor,
                                                        Clazz  referencedClass,
                                                        Member referencedMember)
    {
        int invokeinterfaceConstant =
            (ClassUtil.internalMethodParameterSize(descriptor, false)) << 8;

        return invokeinterface(constantPoolEditor.addInterfaceMethodrefConstant(className,
                                                                                name,
                                                                                descriptor,
                                                                                referencedClass,
                                                                                referencedMember),
                               invokeinterfaceConstant);
    }

    public CompactCodeAttributeComposer invokeinterface(int constantIndex,
                                                        int constant)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKEINTERFACE, constantIndex, constant));
        return this;
    }

    public CompactCodeAttributeComposer invokedynamic(int     bootStrapMethodIndex,
                                                      String  name,
                                                      String  descriptor,
                                                      Clazz[] referencedClasses)
    {
        return invokedynamic(constantPoolEditor.addInvokeDynamicConstant(bootStrapMethodIndex,
                                                                         name,
                                                                         descriptor,
                                                                         referencedClasses));
    }

    public CompactCodeAttributeComposer invokedynamic(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKEDYNAMIC, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer new_(String className)
    {
        return new_(className, null);
    }

    public CompactCodeAttributeComposer new_(String className, Clazz referencedClass)
    {
        return new_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactCodeAttributeComposer new_(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_NEW, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer newarray(int constant)
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_NEWARRAY, constant));
        return this;
    }

    public CompactCodeAttributeComposer anewarray(String className, Clazz referencedClass)
    {
        return anewarray(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactCodeAttributeComposer anewarray(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_ANEWARRAY, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer arraylength()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ARRAYLENGTH));
        return this;
    }

    public CompactCodeAttributeComposer athrow()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_ATHROW));
        return this;
    }

    public CompactCodeAttributeComposer checkcast(String className)
    {
        return checkcast(className, null);
    }

    public CompactCodeAttributeComposer checkcast(String className, Clazz referencedClass)
    {
        return checkcast(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactCodeAttributeComposer checkcast(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_CHECKCAST, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer instanceof_(String className, Clazz referencedClass)
    {
        return instanceof_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactCodeAttributeComposer instanceof_(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_INSTANCEOF, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer monitorenter()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_MONITORENTER));
        return this;
    }

    public CompactCodeAttributeComposer monitorexit()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_MONITOREXIT));
        return this;
    }

    public CompactCodeAttributeComposer wide()
    {
        codeAttributeComposer.appendInstruction(new SimpleInstruction(InstructionConstants.OP_WIDE));
        return this;
    }

    public CompactCodeAttributeComposer multianewarray(String className, Clazz referencedClass)
    {
        return multianewarray(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactCodeAttributeComposer multianewarray(int constantIndex)
    {
        codeAttributeComposer.appendInstruction(new ConstantInstruction(InstructionConstants.OP_MULTIANEWARRAY, constantIndex));
        return this;
    }

    public CompactCodeAttributeComposer ifnull(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFNULL, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer ifnonnull(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_IFNONNULL, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer goto_w(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_GOTO_W, branchOffset));
        return this;
    }

    public CompactCodeAttributeComposer jsr_w(int branchOffset)
    {
        codeAttributeComposer.appendInstruction(new BranchInstruction(InstructionConstants.OP_JSR_W, branchOffset));
        return this;
    }


    // Additional convenience methods.

    /**
     * Pushes the given primitive value on the stack.
     *
     * Operand stack:
     * ... -> ..., value
     *
     * @param primitive the primitive value to be pushed - should never be null.
     * @param internalType      the internal type of the primitive ('Z','B','I',...)
     */
    public void pushPrimitive(Object primitive,
                              char   internalType)
    {
        switch (internalType)
        {
            case TYPE_BOOLEAN:
                if ((Boolean)primitive)
                {
                    iconst_1();
                }
                else
                {
                    iconst_0();
                }
                break;
            case TYPE_BYTE:
                bipush((Byte)primitive);
                break;
            case TYPE_CHAR:
                bipush((Character)primitive);
                break;
            case TYPE_SHORT:
                sipush((Short)primitive);
                break;
            case TYPE_INT:
                ldc((Integer)primitive);
                break;
            case TYPE_LONG:
                ldc2_w((Long)primitive);
                break;
            case TYPE_FLOAT:
                ldc((Float)primitive);
                break;
            case TYPE_DOUBLE:
                ldc2_w((Double)primitive);
        }
    }


    /**
     * Pushes a new array on the stack.
     *
     * Operand stack:
     * ... -> ..., array
     *
     * @param elementType the array element type.
     * @param size        the size of the array to be created.
     */
    public void pushNewArray(String elementType,
                             int    size) {
        // Create new array.
        iconst(size);

        boolean isElementPrimitive = elementType.length() == 1;
        if (isElementPrimitive)
        {
            char primitiveType = elementType.charAt(0);
            newarray(InstructionUtil.arrayTypeFromInternalType(primitiveType));
        }
        else
        {
            anewarray(elementType, null);
        }
    }


    /**
     * Loads the given variable onto the stack.
     *
     * Operand stack:
     * ... -> ..., value
     *
     * @param variableIndex the index of the variable to be loaded.
     * @param internalType  the type of the variable to be loaded.
     */
    public void load(int    variableIndex,
                     String internalType)
    {
        boolean isElementPrimitive = internalType.length() == 1;
        if (isElementPrimitive)
        {
            char primitiveType = internalType.charAt(0);
            loadPrimitive(variableIndex, primitiveType);
        }
        else
        {
            aload(variableIndex);
        }
    }


    /**
     * Loads the given variable of primitive type onto the stack.
     *
     * Operand stack:
     * ... -> ..., value
     *
     * @param variableIndex the index of the variable to be loaded.
     * @param internalType  the primitive type of the variable to be loaded.
     */
    public void loadPrimitive(int  variableIndex,
                              char internalType)
    {
        switch (internalType)
        {
            case TYPE_BOOLEAN:
            case TYPE_BYTE:
            case TYPE_CHAR:
            case TYPE_SHORT:
            case TYPE_INT:
                iload(variableIndex);
                break;
            case TYPE_LONG:
                lload(variableIndex);
                break;
            case TYPE_FLOAT:
                fload(variableIndex);
                break;
            case TYPE_DOUBLE:
                dload(variableIndex);
        }
    }


    /**
     * Stores the value on top of the stack in the variable with given index.
     *
     * Operand stsack:
     * ..., value -> ...
     *
     * @param variableIndex the index of the variable where to store the
     *                      value.
     * @param internalType  the type of the value to be stored.
     */
    public void store(int    variableIndex,
                      String internalType) {
        boolean isElementPrimitive = internalType.length() == 1;
        if (isElementPrimitive)
        {
            char primitiveType = internalType.charAt(0);
            storePrimitive(variableIndex, primitiveType);
        }
        else
        {
            astore(variableIndex);
        }
    }

    /**
     * Stores the primitve value on top of the stack in the variable with given
     * index.
     *
     * Operand stack:
     * ..., value -> ...
     *
     * @param variableIndex the index of the variable where to store the
     *                      value.
     * @param internalType  the primitive type of the value to be stored.
     */
    public void storePrimitive(int  variableIndex,
                               char internalType)
    {
        switch (internalType)
        {
            case TYPE_BOOLEAN:
            case TYPE_BYTE:
            case TYPE_CHAR:
            case TYPE_SHORT:
            case TYPE_INT:
                istore(variableIndex);
                break;
            case TYPE_LONG:
                lstore(variableIndex);
                break;
            case TYPE_FLOAT:
                fstore(variableIndex);
                break;
            case TYPE_DOUBLE:
                dstore(variableIndex);
        }
    }


    /**
     * Stores an element to an array.
     *
     * Operand stack:
     * ..., array, index, value -> ...
     *
     * @param elementType the type of the value to be stored.
     */
    public void storeToArray(String elementType)
    {
        boolean isElementPrimitive = elementType.length() == 1;

        // Store element on stack in array.
        if (isElementPrimitive)
        {
            char primitiveType = elementType.charAt(0);
            switch (primitiveType)
            {
                case TYPE_BOOLEAN:
                case TYPE_BYTE:
                    bastore();
                    break;
                case TYPE_CHAR:
                    castore();
                    break;
                case TYPE_SHORT:
                    sastore();
                    break;
                case TYPE_INT:
                    iastore();
                    break;
                case TYPE_LONG:
                    lastore();
                    break;
                case TYPE_FLOAT:
                    fastore();
                    break;
                case TYPE_DOUBLE:
                    dastore();
                    break;
            }
        }
        else
        {
            aastore();
        }
    }

    /**
     * Appends instructions to print out the given message and the top int on
     * the stack.
     */
    public CompactCodeAttributeComposer appendPrintIntegerInstructions(String message)
    {
        appendPrintInstructions(message);
        appendPrintIntegerInstructions();
        return this;
    }

    /**
     * Appends instructions to print out the given message and the top int on
     * the stack as a hexadecimal value.
     */
    public CompactCodeAttributeComposer appendPrintIntegerHexInstructions(String message)
    {
        appendPrintInstructions(message);
        appendPrintIntegerHexInstructions();
        return this;
    }

    /**
     * Appends instructions to print out the given message and the top long on
     * the stack.
     */
    public CompactCodeAttributeComposer appendPrintLongInstructions(String message)
    {
        appendPrintInstructions(message);
        appendPrintLongInstructions();
        return this;
    }

    /**
     * Appends instructions to print out the given message and the top String on
     * the stack.
     */
    public CompactCodeAttributeComposer appendPrintStringInstructions(String message)
    {
        appendPrintInstructions(message);
        appendPrintStringInstructions();
        return this;
    }

    /**
     * Appends instructions to print out the given message and the top Object on
     * the stack.
     */
    public CompactCodeAttributeComposer appendPrintObjectInstructions(String message)
    {
        appendPrintInstructions(message);
        appendPrintObjectInstructions();
        return this;
    }

    /**
     * Appends instructions to print out the given message and the stack trace
     * of the top Throwable on the stack.
     */
    public CompactCodeAttributeComposer appendPrintStackTraceInstructions(String message)
    {
        appendPrintInstructions(message);
        appendPrintStackTraceInstructions();
        return this;
    }

    /**
     * Appends instructions to print out the given message.
     */
    public CompactCodeAttributeComposer appendPrintInstructions(String message)
    {
        getstatic("java/lang/System", "err", "Ljava/io/PrintStream;");
        ldc(message);
        invokevirtual("java/io/PrintStream", "println", "(Ljava/lang/String;)V");
        return this;
    }

    /**
     * Appends instructions to print out the top int on the stack.
     */
    public CompactCodeAttributeComposer appendPrintIntegerInstructions()
    {
        dup();
        getstatic("java/lang/System", "err", "Ljava/io/PrintStream;");
        swap();
        invokevirtual("java/io/PrintStream", "println", "(I)V");
        return this;
    }

    /**
     * Appends instructions to print out the top integer on the stack as a
     * hexadecimal value.
     */
    public CompactCodeAttributeComposer appendPrintIntegerHexInstructions()
    {
        dup();
        getstatic("java/lang/System", "err", "Ljava/io/PrintStream;");
        swap();
        invokestatic("java/lang/Integer", "toHexString", "(I)Ljava/lang/String;");
        invokevirtual("java/io/PrintStream", "println", "(Ljava/lang/String;)V");
        return this;
    }

    /**
     * Appends instructions to print out the top long on the stack.
     */
    public CompactCodeAttributeComposer appendPrintLongInstructions()
    {
        dup2();
        getstatic("java/lang/System", "err", "Ljava/io/PrintStream;");
        dup_x2();
        pop();
        invokevirtual("java/io/PrintStream", "println", "(J)V");
        return this;
    }

    /**
     * Appends instructions to print out the top String on the stack.
     */
    public CompactCodeAttributeComposer appendPrintStringInstructions()
    {
        dup();
        getstatic("java/lang/System", "err", "Ljava/io/PrintStream;");
        swap();
        invokevirtual("java/io/PrintStream", "println", "(Ljava/lang/String;)V");
        return this;
    }

    /**
     * Appends instructions to print out the top Object on the stack.
     */
    public CompactCodeAttributeComposer appendPrintObjectInstructions()
    {
        dup();
        getstatic("java/lang/System", "err", "Ljava/io/PrintStream;");
        swap();
        invokevirtual("java/io/PrintStream", "println", "(Ljava/lang/Object;)V");
        return this;
    }

    /**
     * Appends instructions to print out the stack trace of the top Throwable
     * on the stack.
     */
    public CompactCodeAttributeComposer appendPrintStackTraceInstructions()
    {
        dup();
        invokevirtual("java/lang/Throwable", "printStackTrace", "()V");
        return this;
    }


    // Implementations for AttributeVisitor.

    public void visitAnyAttribute(Clazz clazz, Attribute attribute) {}


    public void visitCodeAttribute(Clazz clazz, Method method, CodeAttribute codeAttribute)
    {
        codeAttributeComposer.visitCodeAttribute(clazz, method, codeAttribute);
    }


    public static void main(String[] args)
    {
        ProgramClass targetClass = new ProgramClass(0, 0, new Constant[32], 0, 0, 0);

        CompactCodeAttributeComposer composer = new CompactCodeAttributeComposer(targetClass);

        composer.beginCodeFragment(4);
        composer.appendInstruction(0, new SimpleInstruction(InstructionConstants.OP_ICONST_0));
        composer.appendInstruction(1, new VariableInstruction(InstructionConstants.OP_ISTORE, 0));
        composer.appendInstruction(2, new BranchInstruction(InstructionConstants.OP_GOTO, 1));

        composer.beginCodeFragment(4);
        composer.appendInstruction(0, new VariableInstruction(InstructionConstants.OP_IINC, 0, 1));
        composer.appendInstruction(1, new VariableInstruction(InstructionConstants.OP_ILOAD, 0));
        composer.appendInstruction(2, new SimpleInstruction(InstructionConstants.OP_ICONST_5));
        composer.appendInstruction(3, new BranchInstruction(InstructionConstants.OP_IFICMPLT, -3));
        composer.endCodeFragment();

        composer.appendInstruction(3, new SimpleInstruction(InstructionConstants.OP_RETURN));
        composer.endCodeFragment();
    }
}
