/*
 * ProGuard -- shrinking, optimization, obfuscation, and preverification
 *             of Java bytecode.
 *
 * Copyright (c) 2002-2017 Eric Lafortune @ GuardSquare
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package proguard.classfile.editor;

import proguard.classfile.*;
import proguard.classfile.instruction.*;
import proguard.classfile.util.ClassUtil;

import java.util.*;

import static proguard.classfile.ClassConstants.*;

/**
 * This convenience class accumulates simple instructions in a given list,
 * without support for branches or exceptions.
 *
 * @author Johan Leys
 */
public class CompactInstructionListComposer
{
    private final ConstantPoolEditor constantPoolEditor;
    private final List<Instruction>  instructions;


    /**
     * Creates a new CompactInstructionListComposer.
     */
    public CompactInstructionListComposer(ProgramClass targetClass)
    {
        this(new ConstantPoolEditor(targetClass));
    }


    /**
     * Creates a new CompactInstructionListComposer.
     */
    public CompactInstructionListComposer(ConstantPoolEditor constantPoolEditor)
    {
        this(constantPoolEditor, new ArrayList<Instruction>());
    }


    /**
     * Creates a new CompactInstructionListComposer.
     */
    public CompactInstructionListComposer(ProgramClass      targetClass,
                                          List<Instruction> instructions)
    {
        this(new ConstantPoolEditor(targetClass), instructions);
    }


    /**
     * Creates a new CompactInstructionListComposer.
     */
    public CompactInstructionListComposer(ConstantPoolEditor constantPoolEditor,
                                          List<Instruction>  instructions)
    {
        this.constantPoolEditor = constantPoolEditor;
        this.instructions       = instructions;
    }


    /**
     * Resets the internal state of the composer,
     * removing all previously added instructions.
     */
    public void reset()
    {
        instructions.clear();
    }


    /**
     * Appends the given instruction without a defined offset.
     * Branch instructions should have a label, to allow computing the
     * new relative offset.
     * Branch instructions must fit, for instance by enabling automatic
     * shrinking of instructions.
     * @param instruction the instruction to be appended.
     */
    public CompactInstructionListComposer appendInstruction(Instruction instruction)
    {
        instructions.add(instruction);

        return this;
    }

    /**
     * Appends the given instruction without defined offsets.
     * @param instructions the instructions to be appended.
     */
    public void appendInstructions(Instruction[] instructions)
    {
        for (int index = 0; index < instructions.length; index++)
        {
            appendInstruction(instructions[index]);
        }
    }


    /**
     * Returns the built instructions list as an array.
     *
     * @return the built instructions list, as an array.
     */
    public Instruction[] getInstructionsAsArray() {
        Instruction[] instructionsArray = new Instruction[instructions.size()];
        instructions.toArray(instructionsArray);
        return instructionsArray;
    }


    public CompactInstructionListComposer nop()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_NOP));
        return this;
    }

    public CompactInstructionListComposer aconst_null()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ACONST_NULL));
        return this;
    }

    public CompactInstructionListComposer iconst(int constant)
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_0, constant));
        return this;
    }

    public CompactInstructionListComposer iconst_m1()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_M1));
        return this;
    }

    public CompactInstructionListComposer iconst_0()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_0));
        return this;
    }

    public CompactInstructionListComposer iconst_1()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_1));
        return this;
    }

    public CompactInstructionListComposer iconst_2()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_2));
        return this;
    }

    public CompactInstructionListComposer iconst_3()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_3));
        return this;
    }

    public CompactInstructionListComposer iconst_4()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_4));
        return this;
    }

    public CompactInstructionListComposer iconst_5()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ICONST_5));
        return this;
    }

    public CompactInstructionListComposer lconst(int constant)
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LCONST_0, constant));
        return this;
    }

    public CompactInstructionListComposer lconst_0()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LCONST_0));
        return this;
    }

    public CompactInstructionListComposer lconst_1()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LCONST_1));
        return this;
    }

    public CompactInstructionListComposer fconst(int constant)
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCONST_0, constant));
        return this;
    }

    public CompactInstructionListComposer fconst_0()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCONST_0));
        return this;
    }

    public CompactInstructionListComposer fconst_1()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCONST_1));
        return this;
    }

    public CompactInstructionListComposer fconst_2()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCONST_2));
        return this;
    }

    public CompactInstructionListComposer dconst(int constant)
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCONST_0, constant));
        return this;
    }

    public CompactInstructionListComposer dconst_0()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCONST_0));
        return this;
    }

    public CompactInstructionListComposer dconst_1()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCONST_1));
        return this;
    }

    public CompactInstructionListComposer bipush(int constant)
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_BIPUSH, constant));
        return this;
    }

    public CompactInstructionListComposer sipush(int constant)
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_SIPUSH, constant));
        return this;
    }

    public CompactInstructionListComposer ldc(int value)
    {
        return ldc_(constantPoolEditor.addIntegerConstant(value));
    }

    public CompactInstructionListComposer ldc(float value)
    {
        return ldc_(constantPoolEditor.addFloatConstant(value));
    }

    public CompactInstructionListComposer ldc(String string)
    {
        return ldc(string, null, null);
    }

    public CompactInstructionListComposer ldc(Object primitiveArray)
    {
        return ldc_(constantPoolEditor.addPrimitiveArrayConstant(primitiveArray));
    }

    public CompactInstructionListComposer ldc(String string, Clazz referencedClass, Method referencedMember)
    {
        return ldc_(constantPoolEditor.addStringConstant(string, referencedClass, referencedMember));
    }

    public CompactInstructionListComposer ldc(String className, Clazz referencedClass)
    {
        return ldc_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactInstructionListComposer ldc_(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_LDC, constantIndex));
        return this;
    }

    public CompactInstructionListComposer ldc_w(int value)
    {
        return ldc_w_(constantPoolEditor.addIntegerConstant(value));
    }

    public CompactInstructionListComposer ldc_w(float value)
    {
        return ldc_w_(constantPoolEditor.addFloatConstant(value));
    }

    public CompactInstructionListComposer ldc_w(String string)
    {
        return ldc_w(string, null, null);
    }

    public CompactInstructionListComposer ldc_w(String string, Clazz referencedClass, Method referencedMember)
    {
        return ldc_w_(constantPoolEditor.addStringConstant(string, referencedClass, referencedMember));
    }

    public CompactInstructionListComposer ldc_w(String className, Clazz referencedClass)
    {
        return ldc_w_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactInstructionListComposer ldc_w_(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_LDC_W, constantIndex));
        return this;
    }

    public CompactInstructionListComposer ldc2_w(long value)
    {
        return ldc2_w(constantPoolEditor.addLongConstant(value));
    }

    public CompactInstructionListComposer ldc2_w(double value)
    {
        return ldc2_w(constantPoolEditor.addDoubleConstant(value));
    }

    public CompactInstructionListComposer ldc2_w(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_LDC2_W, constantIndex));
        return this;
    }

    public CompactInstructionListComposer iload(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD, variableIndex));
        return this;
    }

    public CompactInstructionListComposer lload(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD, variableIndex));
        return this;
    }

    public CompactInstructionListComposer fload(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD, variableIndex));
        return this;
    }

    public CompactInstructionListComposer dload(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD, variableIndex));
        return this;
    }

    public CompactInstructionListComposer aload(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD, variableIndex));
        return this;
    }

    public CompactInstructionListComposer iload_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD_0));
        return this;
    }

    public CompactInstructionListComposer iload_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD_1));
        return this;
    }

    public CompactInstructionListComposer iload_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD_2));
        return this;
    }

    public CompactInstructionListComposer iload_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ILOAD_3));
        return this;
    }

    public CompactInstructionListComposer lload_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD_0));
        return this;
    }

    public CompactInstructionListComposer lload_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD_1));
        return this;
    }

    public CompactInstructionListComposer lload_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD_2));
        return this;
    }

    public CompactInstructionListComposer lload_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LLOAD_3));
        return this;
    }

    public CompactInstructionListComposer fload_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD_0));
        return this;
    }

    public CompactInstructionListComposer fload_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD_1));
        return this;
    }

    public CompactInstructionListComposer fload_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD_2));
        return this;
    }

    public CompactInstructionListComposer fload_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FLOAD_3));
        return this;
    }

    public CompactInstructionListComposer dload_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD_0));
        return this;
    }

    public CompactInstructionListComposer dload_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD_1));
        return this;
    }

    public CompactInstructionListComposer dload_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD_2));
        return this;
    }

    public CompactInstructionListComposer dload_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DLOAD_3));
        return this;
    }

    public CompactInstructionListComposer aload_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD_0));
        return this;
    }

    public CompactInstructionListComposer aload_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD_1));
        return this;
    }

    public CompactInstructionListComposer aload_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD_2));
        return this;
    }

    public CompactInstructionListComposer aload_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ALOAD_3));
        return this;
    }

    public CompactInstructionListComposer iaload()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IALOAD));
        return this;
    }

    public CompactInstructionListComposer laload()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LALOAD));
        return this;
    }

    public CompactInstructionListComposer faload()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FALOAD));
        return this;
    }

    public CompactInstructionListComposer daload()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DALOAD));
        return this;
    }

    public CompactInstructionListComposer aaload()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_AALOAD));
        return this;
    }

    public CompactInstructionListComposer baload()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_BALOAD));
        return this;
    }

    public CompactInstructionListComposer caload()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_CALOAD));
        return this;
    }

    public CompactInstructionListComposer saload()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_SALOAD));
        return this;
    }

    public CompactInstructionListComposer istore(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE, variableIndex));
        return this;
    }

    public CompactInstructionListComposer lstore(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE, variableIndex));
        return this;
    }

    public CompactInstructionListComposer fstore(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE, variableIndex));
        return this;
    }

    public CompactInstructionListComposer dstore(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE, variableIndex));
        return this;
    }

    public CompactInstructionListComposer astore(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE, variableIndex));
        return this;
    }

    public CompactInstructionListComposer istore_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE_0));
        return this;
    }

    public CompactInstructionListComposer istore_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE_1));
        return this;
    }

    public CompactInstructionListComposer istore_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE_2));
        return this;
    }

    public CompactInstructionListComposer istore_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ISTORE_3));
        return this;
    }

    public CompactInstructionListComposer lstore_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE_0));
        return this;
    }

    public CompactInstructionListComposer lstore_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE_1));
        return this;
    }

    public CompactInstructionListComposer lstore_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE_2));
        return this;
    }

    public CompactInstructionListComposer lstore_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_LSTORE_3));
        return this;
    }

    public CompactInstructionListComposer fstore_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE_0));
        return this;
    }

    public CompactInstructionListComposer fstore_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE_1));
        return this;
    }

    public CompactInstructionListComposer fstore_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE_2));
        return this;
    }

    public CompactInstructionListComposer fstore_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_FSTORE_3));
        return this;
    }

    public CompactInstructionListComposer dstore_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE_0));
        return this;
    }

    public CompactInstructionListComposer dstore_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE_1));
        return this;
    }

    public CompactInstructionListComposer dstore_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE_2));
        return this;
    }

    public CompactInstructionListComposer dstore_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_DSTORE_3));
        return this;
    }

    public CompactInstructionListComposer astore_0()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE_0));
        return this;
    }

    public CompactInstructionListComposer astore_1()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE_1));
        return this;
    }

    public CompactInstructionListComposer astore_2()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE_2));
        return this;
    }

    public CompactInstructionListComposer astore_3()
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_ASTORE_3));
        return this;
    }

    public CompactInstructionListComposer iastore()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IASTORE));
        return this;
    }

    public CompactInstructionListComposer lastore()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LASTORE));
        return this;
    }

    public CompactInstructionListComposer fastore()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FASTORE));
        return this;
    }

    public CompactInstructionListComposer dastore()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DASTORE));
        return this;
    }

    public CompactInstructionListComposer aastore()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_AASTORE));
        return this;
    }

    public CompactInstructionListComposer bastore()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_BASTORE));
        return this;
    }

    public CompactInstructionListComposer castore()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_CASTORE));
        return this;
    }

    public CompactInstructionListComposer sastore()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_SASTORE));
        return this;
    }

    public CompactInstructionListComposer pop()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_POP));
        return this;
    }

    public CompactInstructionListComposer pop2()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_POP2));
        return this;
    }

    public CompactInstructionListComposer dup()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP));
        return this;
    }

    public CompactInstructionListComposer dup_x1()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP_X1));
        return this;
    }

    public CompactInstructionListComposer dup_x2()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP_X2));
        return this;
    }

    public CompactInstructionListComposer dup2()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP2));
        return this;
    }

    public CompactInstructionListComposer dup2_x1()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP2_X1));
        return this;
    }

    public CompactInstructionListComposer dup2_x2()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DUP2_X2));
        return this;
    }

    public CompactInstructionListComposer swap()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_SWAP));
        return this;
    }

    public CompactInstructionListComposer iadd()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IADD));
        return this;
    }

    public CompactInstructionListComposer ladd()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LADD));
        return this;
    }

    public CompactInstructionListComposer fadd()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FADD));
        return this;
    }

    public CompactInstructionListComposer dadd()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DADD));
        return this;
    }

    public CompactInstructionListComposer isub()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ISUB));
        return this;
    }

    public CompactInstructionListComposer lsub()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LSUB));
        return this;
    }

    public CompactInstructionListComposer fsub()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FSUB));
        return this;
    }

    public CompactInstructionListComposer dsub()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DSUB));
        return this;
    }

    public CompactInstructionListComposer imul()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IMUL));
        return this;
    }

    public CompactInstructionListComposer lmul()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LMUL));
        return this;
    }

    public CompactInstructionListComposer fmul()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FMUL));
        return this;
    }

    public CompactInstructionListComposer dmul()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DMUL));
        return this;
    }

    public CompactInstructionListComposer idiv()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IDIV));
        return this;
    }

    public CompactInstructionListComposer ldiv()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LDIV));
        return this;
    }

    public CompactInstructionListComposer fdiv()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FDIV));
        return this;
    }

    public CompactInstructionListComposer ddiv()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DDIV));
        return this;
    }

    public CompactInstructionListComposer irem()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IREM));
        return this;
    }

    public CompactInstructionListComposer lrem()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LREM));
        return this;
    }

    public CompactInstructionListComposer frem()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FREM));
        return this;
    }

    public CompactInstructionListComposer drem()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DREM));
        return this;
    }

    public CompactInstructionListComposer ineg()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_INEG));
        return this;
    }

    public CompactInstructionListComposer lneg()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LNEG));
        return this;
    }

    public CompactInstructionListComposer fneg()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FNEG));
        return this;
    }

    public CompactInstructionListComposer dneg()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DNEG));
        return this;
    }

    public CompactInstructionListComposer ishl()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ISHL));
        return this;
    }

    public CompactInstructionListComposer lshl()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LSHL));
        return this;
    }

    public CompactInstructionListComposer ishr()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ISHR));
        return this;
    }

    public CompactInstructionListComposer lshr()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LSHR));
        return this;
    }

    public CompactInstructionListComposer iushr()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IUSHR));
        return this;
    }

    public CompactInstructionListComposer lushr()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LUSHR));
        return this;
    }

    public CompactInstructionListComposer iand()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IAND));
        return this;
    }

    public CompactInstructionListComposer land()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LAND));
        return this;
    }

    public CompactInstructionListComposer ior()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IOR));
        return this;
    }

    public CompactInstructionListComposer lor()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LOR));
        return this;
    }

    public CompactInstructionListComposer ixor()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IXOR));
        return this;
    }

    public CompactInstructionListComposer lxor()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LXOR));
        return this;
    }

    public CompactInstructionListComposer iinc(int variableIndex,
                                               int constant)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_IINC, variableIndex, constant));
        return this;
    }

    public CompactInstructionListComposer i2l()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2L));
        return this;
    }

    public CompactInstructionListComposer i2f()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2F));
        return this;
    }

    public CompactInstructionListComposer i2d()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2D));
        return this;
    }

    public CompactInstructionListComposer l2i()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_L2I));
        return this;
    }

    public CompactInstructionListComposer l2f()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_L2F));
        return this;
    }

    public CompactInstructionListComposer l2d()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_L2D));
        return this;
    }

    public CompactInstructionListComposer f2i()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_F2I));
        return this;
    }

    public CompactInstructionListComposer f2l()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_F2L));
        return this;
    }

    public CompactInstructionListComposer f2d()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_F2D));
        return this;
    }

    public CompactInstructionListComposer d2i()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_D2I));
        return this;
    }

    public CompactInstructionListComposer d2l()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_D2L));
        return this;
    }

    public CompactInstructionListComposer d2f()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_D2F));
        return this;
    }

    public CompactInstructionListComposer i2b()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2B));
        return this;
    }

    public CompactInstructionListComposer i2c()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2C));
        return this;
    }

    public CompactInstructionListComposer i2s()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_I2S));
        return this;
    }

    public CompactInstructionListComposer lcmp()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LCMP));
        return this;
    }

    public CompactInstructionListComposer fcmpl()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCMPL));
        return this;
    }

    public CompactInstructionListComposer fcmpg()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FCMPG));
        return this;
    }

    public CompactInstructionListComposer dcmpl()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCMPL));
        return this;
    }

    public CompactInstructionListComposer dcmpg()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DCMPG));
        return this;
    }

    public CompactInstructionListComposer ifeq(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFEQ, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ifne(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFNE, branchOffset));
        return this;
    }

    public CompactInstructionListComposer iflt(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFLT, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ifge(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFGE, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ifgt(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFGT, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ifle(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFLE, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ificmpeq(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPEQ, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ificmpne(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPNE, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ificmplt(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPLT, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ificmpge(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPGE, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ificmpgt(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPGT, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ificmple(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFICMPLE, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ifacmpeq(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFACMPEQ, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ifacmpne(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFACMPNE, branchOffset));
        return this;
    }

    public CompactInstructionListComposer goto_(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_GOTO, branchOffset));
        return this;
    }

    public CompactInstructionListComposer jsr(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_JSR, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ret(int variableIndex)
    {
        appendInstruction(new VariableInstruction(InstructionConstants.OP_RET, variableIndex));
        return this;
    }

    public CompactInstructionListComposer tableswitch(int   defaultOffset,
                                                      int   lowCase,
                                                      int   highCase,
                                                      int[] jumpOffsets)
    {
        appendInstruction(new TableSwitchInstruction(InstructionConstants.OP_TABLESWITCH,
                                                     defaultOffset,
                                                     lowCase,
                                                     highCase,
                                                     jumpOffsets));
        return this;
    }

    public CompactInstructionListComposer lookupswitch(int  defaultOffset,
                                                       int[] cases,
                                                       int[] jumpOffsets)
    {
        appendInstruction(new LookUpSwitchInstruction(InstructionConstants.OP_LOOKUPSWITCH,
                                                      defaultOffset,
                                                      cases,
                                                      jumpOffsets));
        return this;
    }

    public CompactInstructionListComposer ireturn()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_IRETURN));
        return this;
    }

    public CompactInstructionListComposer lreturn()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_LRETURN));
        return this;
    }

    public CompactInstructionListComposer freturn()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_FRETURN));
        return this;
    }

    public CompactInstructionListComposer dreturn()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_DRETURN));
        return this;
    }

    public CompactInstructionListComposer areturn()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ARETURN));
        return this;
    }

    public CompactInstructionListComposer return_()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_RETURN));
        return this;
    }

    public CompactInstructionListComposer getstatic(Clazz  referencedClass,
                                                    Member referencedMember)
    {
        return getstatic(referencedClass.getName(),
                         referencedMember.getName(referencedClass),
                         referencedMember.getDescriptor(referencedClass),
                         referencedClass,
                         referencedMember);
    }

    public CompactInstructionListComposer getstatic(String className,
                                                    String name,
                                                    String descriptor)
    {
        return getstatic(className, name, descriptor, null, null);
    }

    public CompactInstructionListComposer getstatic(String className,
                                                    String name,
                                                    String descriptor,
                                                    Clazz  referencedClass,
                                                    Member referencedMember)
    {
        return getstatic(constantPoolEditor.addFieldrefConstant(className,
                                                                name,
                                                                descriptor,
                                                                referencedClass,
                                                                referencedMember));
    }

    public CompactInstructionListComposer getstatic(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_GETSTATIC, constantIndex));
        return this;
    }

    public CompactInstructionListComposer putstatic(Clazz  referencedClass,
                                                    Member referencedMember)
    {
        return putstatic(referencedClass.getName(),
                         referencedMember.getName(referencedClass),
                         referencedMember.getDescriptor(referencedClass),
                         referencedClass,
                         referencedMember);
    }

    public CompactInstructionListComposer putstatic(String className,
                                                    String name,
                                                    String descriptor,
                                                    Clazz  referencedClass,
                                                    Member referencedMember)
    {
        return putstatic(constantPoolEditor.addFieldrefConstant(className,
                                                                name,
                                                                descriptor,
                                                                referencedClass,
                                                                referencedMember));
    }

    public CompactInstructionListComposer putstatic(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_PUTSTATIC, constantIndex));
        return this;
    }

    public CompactInstructionListComposer getfield(String className,
                                                   String name,
                                                   String descriptor)
    {
        return getfield(className, name, descriptor, null, null);
    }

    public CompactInstructionListComposer getfield(String className,
                                                   String name,
                                                   String descriptor,
                                                   Clazz  referencedClass,
                                                   Member referencedMember)
    {
        return getfield(constantPoolEditor.addFieldrefConstant(className,
                                                               name,
                                                               descriptor,
                                                               referencedClass,
                                                               referencedMember));
    }

    public CompactInstructionListComposer getfield(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_GETFIELD, constantIndex));
        return this;
    }

    public CompactInstructionListComposer putfield(String className,
                                                   String name,
                                                   String descriptor,
                                                   Clazz  referencedClass,
                                                   Member referencedMember)
    {
        return putfield(constantPoolEditor.addFieldrefConstant(className,
                                                               name,
                                                               descriptor,
                                                               referencedClass,
                                                               referencedMember));
    }

    public CompactInstructionListComposer putfield(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_PUTFIELD, constantIndex));
        return this;
    }

    public CompactInstructionListComposer invokevirtual(String className,
                                                        String name,
                                                        String descriptor)
    {
        return invokevirtual(className, name, descriptor, null, null);
    }

    public CompactInstructionListComposer invokevirtual(Clazz  referencedClass,
                                                        Member referencedMember)
    {
        return invokevirtual(referencedClass.getName(),
                             referencedMember.getName(referencedClass),
                             referencedMember.getDescriptor(referencedClass),
                             referencedClass,
                             referencedMember);
    }

    public CompactInstructionListComposer invokevirtual(String className,
                                                        String name,
                                                        String descriptor,
                                                        Clazz  referencedClass,
                                                        Member referencedMember)
    {
        return invokevirtual(constantPoolEditor.addMethodrefConstant(className,
                                                                     name,
                                                                     descriptor,
                                                                     referencedClass,
                                                                     referencedMember));
    }

    public CompactInstructionListComposer invokevirtual(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKEVIRTUAL, constantIndex));
        return this;
    }

    public CompactInstructionListComposer invokespecial(String className,
                                                        String name,
                                                        String descriptor)
    {
        return invokespecial(className, name, descriptor, null, null);
    }

    public CompactInstructionListComposer invokespecial(String className,
                                                        String name,
                                                        String descriptor,
                                                        Clazz  referencedClass,
                                                        Member referencedMember)
    {
        return invokespecial(constantPoolEditor.addMethodrefConstant(className,
                                                                     name,
                                                                     descriptor,
                                                                     referencedClass,
                                                                     referencedMember));
    }

    public CompactInstructionListComposer invokespecial(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKESPECIAL, constantIndex));
        return this;
    }

    public CompactInstructionListComposer invokestatic(String className,
                                                       String name,
                                                       String descriptor)
    {
        return invokestatic(className, name, descriptor, null, null);
    }

   public CompactInstructionListComposer invokestatic(Clazz  referencedClass,
                                                      Member referencedMember)
    {
        return invokestatic(referencedClass.getName(),
                            referencedMember.getName(referencedClass),
                            referencedMember.getDescriptor(referencedClass),
                            referencedClass,
                            referencedMember);
    }

    public CompactInstructionListComposer invokestatic(String className,
                                                       String name,
                                                       String descriptor,
                                                       Clazz  referencedClass,
                                                       Member referencedMember)
    {
        return invokestatic(constantPoolEditor.addMethodrefConstant(className,
                                                                    name,
                                                                    descriptor,
                                                                    referencedClass,
                                                                    referencedMember));
    }

    public CompactInstructionListComposer invokestatic(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKESTATIC, constantIndex));
        return this;
    }

    public CompactInstructionListComposer invokeinterface(String className,
                                                          String name,
                                                          String descriptor)
    {
        return invokeinterface(className, name, descriptor, null, null);
    }

    public CompactInstructionListComposer invokeinterface(String className,
                                                          String name,
                                                          String descriptor,
                                                          Clazz  referencedClass,
                                                          Member referencedMember)
    {
        int invokeinterfaceConstant =
            (ClassUtil.internalMethodParameterSize(descriptor, false)) << 8;

        return invokeinterface(constantPoolEditor.addInterfaceMethodrefConstant(className,
                                                                                name,
                                                                                descriptor,
                                                                                referencedClass,
                                                                                referencedMember),
                               invokeinterfaceConstant);
    }

    public CompactInstructionListComposer invokeinterface(int constantIndex,
                                                          int constant)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKEINTERFACE, constantIndex, constant));
        return this;
    }

    public CompactInstructionListComposer invokedynamic(int     bootStrapMethodIndex,
                                                        String  name,
                                                        String  descriptor,
                                                        Clazz[] referencedClasses)
    {
        return invokedynamic(constantPoolEditor.addInvokeDynamicConstant(bootStrapMethodIndex,
                                                                         name,
                                                                         descriptor,
                                                                         referencedClasses));
    }

    public CompactInstructionListComposer invokedynamic(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_INVOKEDYNAMIC, constantIndex));
        return this;
    }

    public CompactInstructionListComposer new_(String className)
    {
        return new_(className, null);
    }

    public CompactInstructionListComposer new_(String className, Clazz referencedClass)
    {
        return new_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactInstructionListComposer new_(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_NEW, constantIndex));
        return this;
    }

    public CompactInstructionListComposer newarray(int constant)
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_NEWARRAY, constant));
        return this;
    }

    public CompactInstructionListComposer anewarray(String className, Clazz referencedClass)
    {
        return anewarray(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactInstructionListComposer anewarray(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_ANEWARRAY, constantIndex));
        return this;
    }

    public CompactInstructionListComposer arraylength()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ARRAYLENGTH));
        return this;
    }

    public CompactInstructionListComposer athrow()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_ATHROW));
        return this;
    }

    public CompactInstructionListComposer checkcast(String className, Clazz referencedClass)
    {
        return checkcast(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactInstructionListComposer checkcast(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_CHECKCAST, constantIndex));
        return this;
    }

    public CompactInstructionListComposer instanceof_(String className, Clazz referencedClass)
    {
        return instanceof_(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactInstructionListComposer instanceof_(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_INSTANCEOF, constantIndex));
        return this;
    }

    public CompactInstructionListComposer monitorenter()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_MONITORENTER));
        return this;
    }

    public CompactInstructionListComposer monitorexit()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_MONITOREXIT));
        return this;
    }

    public CompactInstructionListComposer wide()
    {
        appendInstruction(new SimpleInstruction(InstructionConstants.OP_WIDE));
        return this;
    }

    public CompactInstructionListComposer multianewarray(String className, Clazz referencedClass)
    {
        return multianewarray(constantPoolEditor.addClassConstant(className, referencedClass));
    }

    public CompactInstructionListComposer multianewarray(int constantIndex)
    {
        appendInstruction(new ConstantInstruction(InstructionConstants.OP_MULTIANEWARRAY, constantIndex));
        return this;
    }

    public CompactInstructionListComposer ifnull(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFNULL, branchOffset));
        return this;
    }

    public CompactInstructionListComposer ifnonnull(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_IFNONNULL, branchOffset));
        return this;
    }

    public CompactInstructionListComposer goto_w(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_GOTO_W, branchOffset));
        return this;
    }

    public CompactInstructionListComposer jsr_w(int branchOffset)
    {
        appendInstruction(new BranchInstruction(InstructionConstants.OP_JSR_W, branchOffset));
        return this;
    }


    // Additional convenience methods.

    /**
     * Pushes the given primitive value on the stack.
     *
     * Operand stack:
     * ... -> ..., value
     *
     * @param primitive the primitive value to be pushed - should never be null.
     * @param internalType      the internal type of the primitive ('Z','B','I',...)
     */
    public void pushPrimitive(Object primitive,
                              char   internalType)
    {
        switch (internalType)
        {
            case TYPE_BOOLEAN:
                if ((Boolean)primitive)
                {
                    iconst_1();
                }
                else
                {
                    iconst_0();
                }
                break;
            case TYPE_BYTE:
                bipush((Byte)primitive);
                break;
            case TYPE_CHAR:
                ldc(((Character)primitive).charValue());
                break;
            case TYPE_SHORT:
                sipush((Short)primitive);
                break;
            case TYPE_INT:
                ldc(((Integer)primitive).intValue());
                break;
            case TYPE_LONG:
                ldc2_w((Long)primitive);
                break;
            case TYPE_FLOAT:
                ldc(((Float)primitive).floatValue());
                break;
            case TYPE_DOUBLE:
                ldc2_w((Double)primitive);
        }
    }


    /**
     * Pushes the given int on the stack in the most efficient way
     * (as a bipush, sipush or ldc instruction).
     *
     * @param primitive the primitive value to be pushed.
     */
    public CompactInstructionListComposer pushInt(int primitive)
    {
        if (((byte) primitive) == primitive)
        {
            bipush(primitive);
        }
        else if (((short) primitive) == primitive)
        {
            sipush(primitive);
        }
        else
        {
            ldc(primitive);
        }

        return this;
    }


    /**
     * Pushes a new array on the stack.
     *
     * Operand stack:
     * ... -> ..., array
     *
     * @param elementTypeOrClassName the array element type (or class name in case of objects).
     * @param size                   the size of the array to be created.
     */
    public void pushNewArray(String elementTypeOrClassName,
                             int    size) {
        // Create new array.
        ldc(size);

        boolean isElementPrimitive = elementTypeOrClassName.length() == 1;
        if (isElementPrimitive)
        {
            char primitiveType = elementTypeOrClassName.charAt(0);
            newarray(InstructionUtil.arrayTypeFromInternalType(primitiveType));
        }
        else
        {
            anewarray(elementTypeOrClassName, null);
        }
    }


    /**
     * Loads the given variable onto the stack.
     *
     * Operand stack:
     * ... -> ..., value
     *
     * @param variableIndex the index of the variable to be loaded.
     * @param internalType  the type of the variable to be loaded.
     */
    public void load(int    variableIndex,
                     String internalType)
    {
        boolean isElementPrimitive = internalType.length() == 1;
        if (isElementPrimitive)
        {
            char primitiveType = internalType.charAt(0);
            loadPrimitive(variableIndex, primitiveType);
        }
        else
        {
            aload(variableIndex);
        }
    }


    /**
     * Loads the given variable of primitive type onto the stack.
     *
     * Operand stack:
     * ... -> ..., value
     *
     * @param variableIndex the index of the variable to be loaded.
     * @param internalType  the primitive type of the variable to be loaded.
     */
    public void loadPrimitive(int  variableIndex,
                              char internalType)
    {
        switch (internalType)
        {
            case TYPE_BOOLEAN:
            case TYPE_BYTE:
            case TYPE_CHAR:
            case TYPE_SHORT:
            case TYPE_INT:
                iload(variableIndex);
                break;
            case TYPE_LONG:
                lload(variableIndex);
                break;
            case TYPE_FLOAT:
                fload(variableIndex);
                break;
            case TYPE_DOUBLE:
                dload(variableIndex);
        }
    }


    /**
     * Stores the value on top of the stack in the variable with given index.
     *
     * Operand stsack:
     * ..., value -> ...
     *
     * @param variableIndex the index of the variable where to store the
     *                      value.
     * @param internalType  the type of the value to be stored.
     */
    public void store(int    variableIndex,
                      String internalType) {
        boolean isElementPrimitive = internalType.length() == 1;
        if (isElementPrimitive)
        {
            char primitiveType = internalType.charAt(0);
            storePrimitive(variableIndex, primitiveType);
        }
        else
        {
            astore(variableIndex);
        }
    }

    /**
     * Stores the primitve value on top of the stack in the variable with given
     * index.
     *
     * Operand stack:
     * ..., value -> ...
     *
     * @param variableIndex the index of the variable where to store the
     *                      value.
     * @param internalType  the primitive type of the value to be stored.
     */
    public void storePrimitive(int  variableIndex,
                               char internalType)
    {
        switch (internalType)
        {
            case TYPE_BOOLEAN:
            case TYPE_BYTE:
            case TYPE_CHAR:
            case TYPE_SHORT:
            case TYPE_INT:
                istore(variableIndex);
                break;
            case TYPE_LONG:
                lstore(variableIndex);
                break;
            case TYPE_FLOAT:
                fstore(variableIndex);
                break;
            case TYPE_DOUBLE:
                dstore(variableIndex);
        }
    }


    /**
     * Stores an element to an array.
     *
     * Operand stack:
     * ..., array, index, value -> ...
     *
     * @param elementType the type of the value to be stored.
     */
    public void storeToArray(String elementType)
    {
        boolean isElementPrimitive = elementType.length() == 1;

        // Store element on stack in array.
        if (isElementPrimitive)
        {
            char primitiveType = elementType.charAt(0);
            switch (primitiveType)
            {
                case TYPE_BOOLEAN:
                case TYPE_BYTE:
                    bastore();
                    break;
                case TYPE_CHAR:
                    castore();
                    break;
                case TYPE_SHORT:
                    sastore();
                    break;
                case TYPE_INT:
                    iastore();
                    break;
                case TYPE_LONG:
                    lastore();
                    break;
                case TYPE_FLOAT:
                    fastore();
                    break;
                case TYPE_DOUBLE:
                    dastore();
                    break;
            }
        }
        else
        {
            aastore();
        }
    }
}
